import logging
import os

import colorlog

logger = logging.getLogger("hallux_logger")

log_level = os.environ.get("LOG_LEVEL", logging.WARNING)

logger.setLevel(log_level)

# Create a console handler
handler = logging.StreamHandler()
handler.setLevel(log_level)

# Create a formatter
formatter = colorlog.ColoredFormatter(
    "%(log_color)s%(levelname)-8s%(reset)s %(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "MESSAGE": "black",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    },
    secondary_log_colors={},
    style="%",
)

# Add the formatter to the handler
handler.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(handler)


def message(msg: str, *args, **kwargs):
    print(msg, *args, **kwargs)


logger.message = message


def log_multiline(header: str, content: any, level: str = "debug"):
    getattr(logger, level)("")
    getattr(logger, level)(header)
    getattr(logger, level)("=" * 40)
    getattr(logger, level)("")

    if content is None:
        return

    if isinstance(content, str):
        for line in content.split("\n"):
            getattr(logger, level)(line)
    else:
        getattr(logger, level)(str(content))


logger.log_multiline = log_multiline

# Now you can log messages with various severity levels like this:
# logger.debug("debug message")
# logger.info("info message")
# logger.warning("warning message")
# logger.error("error message")
# logger.critical("critical message")



# Copyright: Hallux team, 2023 - 2024

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Final

from ..proposals.proposal_engine import ProposalEngine
from .annotations import get_language


class IssueDescriptor(ABC):
    def __init__(
        self,
        tool: str,
        filename: str,
        issue_line: int = 0,
        description: str = "",
        issue_type: str = "warning",
        language: str | None = None,
        comment: str | None = None,
        line_comment: str | None = None,
    ):
        self.tool: Final[str] = tool
        self.filename: Final[str] = filename
        self.issue_line: int = issue_line
        self.description: str = description
        self.message_lines: list[str] = []
        self.issue_type: str = issue_type
        self.line_comment = line_comment

        if language is None or comment is None:
            language, comment = get_language(filename)
        self.language: Final[str] = language
        self.comment: Final[str] = comment

        if line_comment is None and comment is not None:
            # TODO: paarametrize line_comment template
            self.line_comment = " " + comment + " <-- fix around this line " + str(issue_line)

    @abstractmethod
    def list_proposals(self) -> ProposalEngine:
        # ToDo: we can provide QueryBackend or even a list[QueryBackend] in here
        # (or some kind of prioritized container)
        # This might give ability to try different Backends for different Proposals
        pass



# Copyright: Hallux team, 2023 - 2024

from __future__ import annotations

import ast
import copy

from ..issues.issue import IssueDescriptor
from .simple_proposal import SimpleProposal


class PythonProposal(SimpleProposal):
    def __init__(
        self,
        issue: IssueDescriptor,
        # radius or tuple with [start_line, end_line]
        radius_or_range: int | tuple[int, int] = 4,
        extract_function: bool = False,
    ):
        super().__init__(issue, radius_or_range)
        if extract_function:
            self.extract_function(issue)
        self.code_offset = 50000
        self.count_code_offset()
        if self.code_offset > 0:
            self.remove_code_offset()

    def extract_function(self, issue):
        with open(issue.filename, "rt") as f:
            python_source = f.read()
        parsed_ast = ast.parse(python_source, filename=issue.filename)
        target_function = self.find_target_function(parsed_ast, issue.issue_line)
        if target_function:
            func_range = (target_function.lineno, target_function.end_lineno)
            self._set_code_range(func_range)

    def find_target_function(self, parsed_ast, issue_line):
        for node in ast.walk(parsed_ast):
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                if node.lineno <= issue_line <= node.end_lineno:
                    return node

    def count_code_offset(self):
        for line in self.issue_lines:
            lsline = line.lstrip(" ")
            if len(lsline) > 0 and not (lsline == "\n" or lsline == "\r\n"):
                self.code_offset = min(self.code_offset, len(line) - len(lsline))

    def remove_code_offset(self):
        for i, line in enumerate(self.issue_lines):
            self.issue_lines[i] = line[self.code_offset :]
        self.proposed_lines = copy.deepcopy(self.issue_lines)

    def _merge_lines(self, proposed_lines: list[str]) -> bool:
        if self.code_offset > 0:
            offset = " " * self.code_offset

            for i, line in enumerate(proposed_lines):
                if len(line) > 0:
                    proposed_lines[i] = offset + line

            for i, line in enumerate(self.issue_lines):
                if len(line) > 0:
                    self.issue_lines[i] = offset + line

        return super()._merge_lines(proposed_lines)



# Copyright: Hallux team, 2023-2024

from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Final

import requests

from hallux.logger import logger

from ...backends.query_backend import QueryBackend
from ...targets.diff import DiffTarget
from ..issue_solver import IssueSolver
from .issue import IssueDescriptor, SonarIssue


class OverrideQueryBackend(QueryBackend):
    """
    With SonarQube we only may solve 1 issue per file.
    We override QueryBackend to track files with successful fixes.
    Later, this file list is used to avoid any other fixes
    """

    def __init__(self, underlying_backend: QueryBackend, already_fixed_files: list[str]):
        super().__init__()
        self.backend: Final[QueryBackend] = underlying_backend
        self.already_fixed_files: Final[list[str]] = already_fixed_files

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        return self.backend.query(request, issue, issue_lines)

    def report_successful_fix(self, issue, proposal) -> None:
        self.already_fixed_files.append(issue.filename)
        self.backend.report_successful_fix(issue, proposal)


class Sonar_IssueSolver(IssueSolver):
    SONAR_TOKEN: Final[str] = "SONAR_TOKEN"

    def __init__(
        self,
        config_path: Path,
        run_path: Path,
        command_dir: str = ".",
        validity_test: str | None = None,
        url: str | None = None,
        token: str | None = None,
        project: str | None = None,
        search_params: str = "resolved=false&severities=MINOR,MAJOR,CRITICAL&statuses=OPEN,CONFIRMED,REOPENED",
        argvalue: str | None = None,  # example: "branch=65-sonar-fixes"
    ):
        """
        IssueSolver implementation for SonarQube
        :param config_path: Directory with .hallux file
        :param run_path: Directory, where hallux was run
        :param command_dir: Directory, passed to hallux in CLI
        :param validity_test: script
        :param url:
        :param token:
        :param project:
        :param search_params:
        :param argvalue: It could be a path to a .json file, or a string with extra params
        """
        super().__init__(config_path, run_path, command_dir, validity_test=validity_test)

        self.token: Final[str | None] = token if token is not None else os.getenv(self.SONAR_TOKEN)
        self.url: Final[str | None] = url
        self.project: Final[str | None] = project
        self.search_params: Final[str] = search_params
        self.argvalue: Final[str | None] = argvalue

        # With SonarQube we only may solve 1 issue per file.
        self.already_fixed_files: Final[list[str]] = []

    def solve_issues(self, diff_target: DiffTarget, query_backend: QueryBackend):
        if not self._check_file() and (not self.token or not self.url or not self.project):
            logger.error("SonarQube: token, url or project not configured")
            return

        logger.message("Process SonarQube issues:")
        new_query_backend = OverrideQueryBackend(query_backend, self.already_fixed_files)
        super().solve_issues(diff_target, new_query_backend)

    def _check_file(self):
        return self.argvalue and self.argvalue.endswith(".json") and Path(self.argvalue).exists()

    def is_issue_fixed(self) -> bool:
        if self.validity_test is None:
            return True
        else:
            return super().is_issue_fixed()

    def list_issues(self) -> list[IssueDescriptor]:
        issues: list[IssueDescriptor] = []

        # ToDo: make some caching for issues, in order not to query sonar every time
        response_json: str
        if self._check_file():
            logger.info("SONAR extra_params is a json file: " + self.argvalue)
            with open(self.argvalue, "r") as f:
                response_json = f.read()
        else:
            try:
                request: str = self.url + "/api/issues/search?" + self.search_params
                if self.argvalue is not None:
                    request += "&" + self.argvalue
                if self.project is not None:
                    request += "&componentKeys=" + self.project
                logger.debug("Sonar URL: " + request)
                x = requests.get(url=request, headers={"Authorization": "Bearer " + self.token})
            except Exception:
                return []

            if x.status_code == 200:
                response_json = x.text
            else:
                logger.error(f'SonarQube: error "{x.status_code}" while making request to {request}')
                return []

        logger.log_multiline("SonarQube Issues:", self._format_issues(response_json))

        issues.extend(SonarIssue.parseIssues(response_json, self.already_fixed_files))

        return issues

    def _format_issues(self, issues):
        issues_data = json.loads(issues)
        table_format = "{:<60} {:<10} {:<60}"
        response = table_format.format("Message", "Severity", "Component")
        response += "\n" + ("-" * 130)
        for issue in issues_data.get("issues", []):
            response += "\n"
            response += table_format.format(
                issue["message"][:60], issue["severity"], issue["component"].split(":")[-1][:60]
            )
        return response



# Copyright: Hallux team, 2023

from __future__ import annotations

import difflib
from abc import ABC, abstractmethod
from typing import Any


class DiffProposal(ABC):
    # This class does not contain any fancy members like IssueDescriptor to avoid circular imports
    def __init__(
        self, filename: str = "", description: str = "", issue_line: int = 0, start_line: int = 0, end_line: int = 0
    ):
        self.filename: str = filename
        self.description: str = description
        self.issue_line: int = issue_line
        self.start_line: int = start_line
        self.end_line: int = end_line
        self.all_lines: list[str] = []
        self.issue_lines: list[str] = []
        self.proposed_lines: list[str] = []

    @abstractmethod
    def try_fixing(self, query_backend, diff_target) -> bool:
        pass

    def try_fixing_with_priority(self, query_backend, diff_target, used_backend) -> tuple[bool, Any]:
        previous_backend = query_backend.previous_backend()

        if previous_backend is not None and previous_backend != used_backend:
            return self.try_fixing_with_priority(previous_backend, diff_target, used_backend)
        else:
            return self.try_fixing(query_backend, diff_target), query_backend

    def print_diff(self, lines1, lines2):
        diff = difflib.unified_diff(lines1, lines2, fromfile=self.filename, tofile=self.filename)
        for line in diff:
            if line.startswith("-"):
                print(f"\033[91m{line}\033[0m", end="")
            elif line.startswith("+"):
                print(f"\033[92m{line}\033[0m", end="")
            else:
                print(line, end="")
        print()



# Copyright: Hallux team, 2023

from __future__ import annotations

from pathlib import Path

from ..auxilary import set_directory
from ..proposals.diff_proposal import DiffProposal
from .diff import DiffTarget


# Saves diff-proposals directly into local filesystem
class FilesystemTarget(DiffTarget):
    def __init__(self):
        self.existing_proposal: DiffProposal | None = None
        self.base_path: Path | None = None

    def apply_diff(self, proposal: DiffProposal) -> bool:
        if self.existing_proposal is not None:
            raise SystemError("FilesystemTarget: Cannot apply new diff, before last one committed or reverted")

        if not Path(proposal.filename).exists():
            raise SystemError(f"FilesystemTarget: Cannot find file: {proposal.filename}")

        self.base_path = Path().absolute()
        self.existing_proposal = proposal  # save new diff in the memory

        with open(proposal.filename, "wt") as file:
            for line in range(0, proposal.start_line - 1):
                file.write(proposal.all_lines[line])

            for code_line in proposal.proposed_lines:
                file.write(code_line)

            for line in range(proposal.end_line, len(proposal.all_lines)):
                file.write(proposal.all_lines[line])
                # if line < len(proposal.all_lines) - 1:
                #     file.write("\n")  #ToDo: do we need it now?

            file.close()
        return True

    def revert_diff(self) -> None:
        if self.existing_proposal is not None:
            with set_directory(self.base_path):
                with open(self.existing_proposal.filename, "wt") as file:
                    all_lines = self.existing_proposal.all_lines
                    for line in range(len(all_lines)):
                        file.write(all_lines[line])
                        # if line < len(all_lines) - 1:
                        #     file.write("\n") #ToDo: do we need it now?

            self.existing_proposal = None
            self.base_path = None

    def commit_diff(self) -> bool:
        self.existing_proposal = None
        self.base_path = None
        return True

    def requires_refresh(self) -> bool:
        return True



# Copyright: Hallux team, 2023 - 2024

from __future__ import annotations

from ...issues.issue import IssueDescriptor
from ...proposals.proposal_engine import ProposalEngine, ProposalList
from ...proposals.python_proposal import PythonProposal


class PythonIssue(IssueDescriptor):
    def __init__(self, filename: str, issue_line: int = 0, description: str = "", tool: str = "ruff"):
        super().__init__(
            language="python", tool=tool, filename=filename, issue_line=issue_line, description=description
        )

    def list_proposals(self) -> ProposalEngine:
        return ProposalList(
            [
                PythonProposal(self, extract_function=True),
                PythonProposal(self, radius_or_range=4),
                PythonProposal(self, radius_or_range=6),
            ]
        )

    @staticmethod
    def parseIssues(ruff_output: str, tool: str = "ruff", keyword: str = "") -> list[PythonIssue]:
        issues: list[PythonIssue] = []
        warnings: list[str] = ruff_output.split("\n")
        for warn in warnings[:-2]:
            warn_arr = warn.split(" ")
            if len(warn_arr) < 3:
                break
            filename_line_col = warn_arr[0].split(":")
            if len(filename_line_col) < 3:
                break
            filename = filename_line_col[0]
            issue_line = int(filename_line_col[1])

            if len(keyword) == 0 or str(warn_arr[1:]).startswith(keyword):
                issue = PythonIssue(
                    filename=filename,
                    issue_line=issue_line,
                    description=str(" ".join(warn_arr[1:]).lstrip(" ")),
                    tool=tool,
                )
                issues.append(issue)

        return issues



# Copyright: Hallux team, 2023

from __future__ import annotations

import hashlib
import json
from pathlib import Path

from hallux.models import PromptConfig

from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor
from ..proposals.diff_proposal import DiffProposal


class DummyBackend(QueryBackend):
    def __init__(
        self,
        filename: str,
        base_path: Path,
        type="dummy",
        previous_backend: QueryBackend | None = None,
        prompt: PromptConfig = PromptConfig(),
    ):
        assert type == "dummy"
        super().__init__(base_path, previous_backend)

        self.prompt = prompt
        self.filename: str
        self.was_modified: bool = False
        if Path(filename).is_absolute():
            self.filename = filename
        else:
            self.filename = str(self.base_path.joinpath(filename))

        if Path(self.filename).exists():
            with open(self.filename, "rt") as file:
                self.json = json.load(file)
        else:
            self.json = {}
            self.save_on_exit = True

    def issue_hash(self, description: str, issue_lines: list[str]) -> str:
        encoded_data = str(description + "\n" + "\n".join(issue_lines)).encode("utf8")
        return str(hashlib.md5(encoded_data).hexdigest())

    def __del__(self):
        if self.was_modified:
            with open(self.filename, "wt") as file:
                json.dump(self.json, file)

    def report_successful_fix(self, issue: IssueDescriptor, proposal: DiffProposal) -> None:
        hash = self.issue_hash(issue.description, proposal.issue_lines)
        if hash not in self.json:
            self.json[hash] = "\n".join(proposal.proposed_lines)

        self.was_modified = True

        if self.previous is not None:
            self.previous.report_successful_fix(issue, proposal)

    def query(
        self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] | None = None
    ) -> list[str]:
        if issue is not None and issue_lines is not None:
            hash = self.issue_hash(issue.description, issue_lines)
            return self.json.get(hash, [])

        return []



# Copyright: Hallux team, 2024
from __future__ import annotations


def get_language(filename: str) -> tuple[str, str]:
    ext = filename.split(".")[-1]
    language = ext
    comment_str = "//"

    if ext in ["py"]:
        return "python", "#"

    if ext in ["cpp", "hpp", "c", "h"]:
        return "cpp", comment_str

    if ext in ["js", "jsx", "mjs"]:
        return "javascript", comment_str

    if ext in ["ts", "tsx"]:
        return "typescript", comment_str

    if ext in ["java"]:
        return "java", comment_str

    if ext in ["swift"]:
        return "swift", comment_str

    # TODO: shall we default comment_char?
    return language, None
